from dataclasses import dataclass
from typing import Dict
from typing import Optional

from discord import Client as DiscordClient
from discord import Guild, VoiceChannel, CategoryChannel
from discord import Role, PermissionOverwrite

from discord_coworking.command.api import Result, UnundoableResult
from discord_coworking.command.guild import GuildResult, GuildCommand, GuildPredicate
from discord_coworking.command.overwrites import OverwritesMixin
from discord_coworking.command.predicate import ANY, NONE
from discord_coworking.command.predicate import ByAttributesPredicate, Predicate


class VoiceChannelPredicate(ByAttributesPredicate[VoiceChannel]):

    @classmethod
    def of(cls, voice_channel: VoiceChannel):
        return cls.create(id=voice_channel.id)


@dataclass()
class VoiceChannelCreated(GuildResult):
    voice_channel: Predicate[VoiceChannel] = ANY

    @classmethod
    def create(cls, guild: Guild, voice_channel: VoiceChannel):
        return cls(
            guild=GuildPredicate.of(guild),
            voice_channel=VoiceChannelPredicate.of(voice_channel),
        )

    async def undo_on_guild(self, client: DiscordClient, guild: Guild):
        for voice_channel in self.voice_channel.filter(guild.voice_channels):
            await voice_channel.delete()


@dataclass()
class CreateVoiceChannel(GuildCommand, OverwritesMixin):
    name: str = None
    bitrate: Optional[int] = None
    user_limit: Optional[int] = None
    category: Predicate[CategoryChannel] = NONE
    overwrites: Optional[Dict[Predicate[Role], PermissionOverwrite]] = None

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> VoiceChannelCreated:
        voice_channel = await guild.create_voice_channel(
            name=self.name,
            bitrate=self.bitrate,
            user_limit=self.user_limit,
            category=self.category.any(guild.categories),
            overwrites=self.get_overwrites(guild),
        )
        return VoiceChannelCreated.create(guild, voice_channel)


class DeleteVoiceChannel(GuildCommand):
    voice_channel: Predicate[VoiceChannel] = ANY

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> Result:
        for voice_channel in self.voice_channel.filter(guild.voice_channels):
            await voice_channel.delete()
        return UnundoableResult.create(self)



import functools
from dataclasses import dataclass

from discord import Client as DiscordClient

from discord_coworking.command.api import UndoableResult, Result, Command
from discord_coworking.command.batch import BatchResult


@dataclass()
class UndoableResultImpl(UndoableResult):
    result: Result = None

    async def undo(self, client: DiscordClient):
        self.logger.debug(f'Undoing command {self.command}')
        return await self.result.undo(client)

    @classmethod
    def create(cls, command: Command, result: Result):
        return cls(
            command=command,
            result=result,
        )


def async_batch(fn):
    @functools.wraps(fn)
    async def wrapper(command, *args, **kwargs) -> Result:
        results = []
        try:
            async for result in fn(command, *args, **kwargs):
                results.append(result)
        except Exception as ex:
            client = kwargs.get('client', None)
            if client is None:
                client = args[0]
            for result in results[::-1]:
                await result.undo(client)
            raise ex
        return UndoableResultImpl.create(command, BatchResult(results[::-1]))

    return wrapper



import emoji
from datetime import datetime
import asyncio
from discord import Message, Client
import os
import docker
from docker.types import Mount

import re

from discord_coworking.bot import MessageHandler
from dataclasses import dataclass, field


@dataclass()
class DockerRun:
    language: str
    code: str
    docker_client: docker.DockerClient = field(default_factory=docker.from_env)

    loading_emoji = emoji.EMOJI_ALIAS_UNICODE[':hourglass:']
    success_emoji = emoji.EMOJI_ALIAS_UNICODE[':white_check_mark:']
    error_emoji = emoji.EMOJI_ALIAS_UNICODE[':x:']
    default_emoji = emoji.EMOJI_ALIAS_UNICODE[':gear:']

    emoji = {
        'python': emoji.EMOJI_ALIAS_UNICODE[':snake:'],
        'java': emoji.EMOJI_ALIAS_UNICODE[':coffee:'],
        'js': emoji.EMOJI_ALIAS_UNICODE[':poop:'],
        'ts': emoji.EMOJI_ALIAS_UNICODE[':sauropod:'],
        'php': emoji.EMOJI_ALIAS_UNICODE[':elephant:'],
        'c': emoji.EMOJI_ALIAS_UNICODE[':regional_indicator_c:'],
        'cpp': emoji.EMOJI_ALIAS_UNICODE[':rat:'],
        'rust': emoji.EMOJI_ALIAS_UNICODE[':crab:'],
        'sh': emoji.EMOJI_ALIAS_UNICODE[':shell:'],
        'bash': emoji.EMOJI_ALIAS_UNICODE[':computer:'],
        'go': emoji.EMOJI_ALIAS_UNICODE[':chipmunk:'],
        'lua': emoji.EMOJI_ALIAS_UNICODE[':new_moon:'],
        'ruby': emoji.EMOJI_ALIAS_UNICODE[':diamonds:'],
        'lisp': emoji.EMOJI_ALIAS_UNICODE[':alien:'],
        'perl': emoji.EMOJI_ALIAS_UNICODE[':onion:'],
        'r': emoji.EMOJI_ALIAS_UNICODE[':regional_indicator_r:'],
        'julia': emoji.EMOJI_ALIAS_UNICODE[':regional_indicator_j:'],
        'kotlin': emoji.EMOJI_ALIAS_UNICODE[':regional_indicator_k:'],
    }

    aliases = {
        'c++': 'cpp',
        'javascript': 'js',
        'typescript': 'ts',
    }

    images = {
        'python': 'python:3.10',
        'java': 'openjdk:11',
        'js': 'node:lts',
        'ts': 'hayd/deno',
        'php': 'php',
        'c': 'gcc',
        'cpp': 'gcc',
        'rust': 'rust',
        'sh': 'fedora',
        'bash': 'fedora',
        'go': 'golang',
        'lua': 'woahbase/alpine-lua',
        'ruby': 'ruby',
        'haskell': 'haskell',
        'lisp': 'daewok/lisp-devel',
        'perl': 'perl',
        'r': 'r-base',
        'julia': 'julia',
        'kotlin': 'zenika/kotlin',
    }

    commands = {
        'python': 'python {}',
        'java': 'java {}',
        'js': 'node {}',
        'ts': 'deno run -q {}',
        'php': 'php {}',
        'c': '/bin/bash -c "gcc {} -o /tmp/a.out;/tmp/a.out"',
        'cpp': '/bin/bash -c "gcc -x c++ {} -lstdc++ -o /tmp/a.out;/tmp/a.out"',
        'rust': '/bin/bash -c "rustc {} -o /tmp/a.out;/tmp/a.out"',
        'sh': '/bin/sh {}',
        'bash': '/bin/bash {}',
        'go': 'go run {}',
        'lua': 'lua {}',
        'ruby': 'ruby {}',
        'haskell': 'runghc {}',
        'lisp': 'sbcl --script {}',
        'perl': 'perl {}',
        'r': 'Rscript {}',
        'julia': 'julia {}',
        'kotlin': 'kotlinc -script  {}',
    }

    scripts = {
        'python': 'script.py',
        'java': 'Script.java',
        'js': 'script.js',
        'ts': 'script.ts',
        'php': 'script.php',
        'c': 'script.c',
        'cpp': 'script.cpp',
        'rust': 'script.rs',
        'sh': 'script.sh',
        'bash': 'script.sh',
        'go': 'script.go',
        'lua': 'script.lua',
        'ruby': 'script.rb',
        'haskell': 'script.hs',
        'lisp': 'script.lisp',
        'perl': 'script.pl',
        'r': 'script.r',
        'julia': 'julia.jl',
        'kotlin': 'script.kts',
    }

    async def __call__(self, client: Client, message: Message):
        try:
            language = self.aliases.get(self.language, self.language)
            image = self.images.get(language)
            command = self.commands.get(language)
            script = self.scripts.get(language)
            if None in [image, command, script]:
                return
            await message.add_reaction(self.emoji.get(language, self.default_emoji))
            volume = f'/tmp/discord-coworking-{message.id}/'
            os.mkdir(volume)
            with open(os.path.join(volume, script), 'w+') as output:
                output.write(self.code)
            container = self.docker_client.containers.run(
                image,
                command.format(os.path.join(volume, script)),
                auto_remove=False,
                mounts=[
                    Mount(source=volume, target=volume, type='bind')
                ],
                detach=True,
                network_mode='none',
                mem_limit='256m',
                nano_cpus=int(0.25 * 10 ** 9),
            )
            await message.add_reaction(self.loading_emoji)
            started_at = datetime.now()
            running = True
            while running:
                try:
                    result = container.wait(timeout=0.01)
                    running = False
                except:
                    diff = datetime.now() - started_at
                    if diff.seconds >= 10:
                        container.remove(force=True)
                        raise Exception('Timeout')
                    await asyncio.sleep(0.5)
            stdout = container.logs()
            container.remove()
            if result['StatusCode'] != 0:
                raise Exception(stdout.decode())
            await message.add_reaction(self.success_emoji)
            await message.reply(stdout.decode())
        except Exception as ex:
            await message.add_reaction(self.error_emoji)
            await message.reply(f'```{ex.args[0]}```')
        finally:
            await message.remove_reaction(self.loading_emoji, client.user)


class RunCodeInDocker(MessageHandler):
    pattern = re.compile('```(?P<language>[a-zA-Z0-9+#-]+)\\n(?P<code>[^`]*)```')

    def accept(self, message: Message) -> bool:
        match = self.pattern.findall(message.clean_content)
        return bool(match)

    async def handle(self, client: Client, message: Message):
        for match in self.pattern.findall(message.clean_content):
            language, code = match
            run = DockerRun(language=language.lower(), code=code)
            asyncio.ensure_future(run(client, message))



from dataclasses import dataclass, field

from discord import Client as DiscordClient, Guild, PermissionOverwrite

from discord_coworking.command.api import Result
from discord_coworking.command.category import CreateCategory
from discord_coworking.command.decorator import async_batch
from discord_coworking.command.guild import GuildCommand
from discord_coworking.command.role import CreateRole, DEFAULT_ROLE
from discord_coworking.command.text_channel import CreateTextChannel
from discord_coworking.command.voice_channel import CreateVoiceChannel

NO_PERMISSIONS = PermissionOverwrite(
    create_instant_invite=False,
    kick_members=False,
    ban_members=False,
    administrator=False,
    manage_channels=False,
    manage_guild=False,
    add_reactions=False,
    view_audit_log=False,
    priority_speaker=False,
    stream=False,
    read_messages=False,
    view_channel=False,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=False,
    mention_everyone=False,
    external_emojis=False,
    use_external_emojis=False,
    view_guild_insights=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    change_nickname=False,
    manage_nicknames=False,
    manage_roles=False,
    manage_permissions=False,
    manage_webhooks=False,
    manage_emojis=False,
)


class OpenOrganization:
    DEFAULT_OTHERS_PERMISSIONS = PermissionOverwrite(
        create_instant_invite=False,
        kick_members=False,
        ban_members=False,
        administrator=False,
        manage_channels=False,
        manage_guild=False,
        add_reactions=False,
        view_audit_log=False,
        priority_speaker=False,
        stream=True,
        read_messages=True,
        view_channel=True,
        send_messages=True,
        send_tts_messages=False,
        manage_messages=False,
        embed_links=True,
        attach_files=True,
        read_message_history=True,
        mention_everyone=False,
        external_emojis=False,
        use_external_emojis=False,
        view_guild_insights=False,
        connect=True,
        speak=True,
        mute_members=False,
        deafen_members=False,
        move_members=False,
        use_voice_activation=True,
        change_nickname=True,
        manage_nicknames=False,
        manage_roles=False,
        manage_permissions=False,
        manage_webhooks=False,
        manage_emojis=False,
    )

    DEFAULT_FORMER_PERMISSIONS = DEFAULT_OTHERS_PERMISSIONS

    DEFAULT_ADMIN_PERMISSIONS = PermissionOverwrite(
        create_instant_invite=True,
        kick_members=True,
        ban_members=False,
        administrator=False,
        manage_channels=True,
        manage_guild=False,
        add_reactions=True,
        view_audit_log=True,
        priority_speaker=True,
        stream=True,
        read_messages=True,
        view_channel=True,
        send_messages=True,
        send_tts_messages=True,
        manage_messages=True,
        embed_links=True,
        attach_files=True,
        read_message_history=True,
        mention_everyone=True,
        external_emojis=True,
        use_external_emojis=True,
        view_guild_insights=False,
        connect=True,
        speak=True,
        mute_members=True,
        deafen_members=True,
        move_members=True,
        use_voice_activation=True,
        change_nickname=True,
        manage_nicknames=False,
        manage_roles=False,
        manage_permissions=False,
        manage_webhooks=False,
        manage_emojis=False,
    )

    DEFAULT_ASSOCIATE_PERMISSIONS = PermissionOverwrite(
        create_instant_invite=True,
        kick_members=False,
        ban_members=False,
        administrator=False,
        manage_channels=False,
        manage_guild=False,
        add_reactions=True,
        view_audit_log=True,
        priority_speaker=False,
        stream=True,
        read_messages=True,
        view_channel=True,
        send_messages=True,
        send_tts_messages=True,
        manage_messages=True,
        embed_links=True,
        attach_files=True,
        read_message_history=True,
        mention_everyone=True,
        external_emojis=True,
        use_external_emojis=True,
        view_guild_insights=False,
        connect=True,
        speak=True,
        mute_members=True,
        deafen_members=False,
        move_members=False,
        use_voice_activation=True,
        change_nickname=True,
        manage_nicknames=False,
        manage_roles=False,
        manage_permissions=False,
        manage_webhooks=False,
        manage_emojis=False,
    )


class PrivateOrganization:
    DEFAULT_OTHERS_PERMISSIONS = PermissionOverwrite(
        create_instant_invite=False,
        kick_members=False,
        ban_members=False,
        administrator=False,
        manage_channels=False,
        manage_guild=False,
        add_reactions=False,
        view_audit_log=False,
        priority_speaker=False,
        stream=False,
        read_messages=False,
        view_channel=False,
        send_messages=False,
        send_tts_messages=False,
        manage_messages=False,
        embed_links=False,
        attach_files=False,
        read_message_history=False,
        mention_everyone=False,
        external_emojis=False,
        use_external_emojis=False,
        view_guild_insights=False,
        connect=False,
        speak=False,
        mute_members=False,
        deafen_members=False,
        move_members=False,
        use_voice_activation=False,
        change_nickname=False,
        manage_nicknames=False,
        manage_roles=False,
        manage_permissions=False,
        manage_webhooks=False,
        manage_emojis=False,
    )

    DEFAULT_ADMIN_PERMISSIONS = PermissionOverwrite(
        create_instant_invite=True,
        kick_members=True,
        ban_members=False,
        administrator=False,
        manage_channels=True,
        manage_guild=False,
        add_reactions=True,
        view_audit_log=True,
        priority_speaker=True,
        stream=True,
        read_messages=True,
        view_channel=True,
        send_messages=True,
        send_tts_messages=True,
        manage_messages=True,
        embed_links=True,
        attach_files=True,
        read_message_history=True,
        mention_everyone=True,
        external_emojis=True,
        use_external_emojis=True,
        view_guild_insights=False,
        connect=True,
        speak=True,
        mute_members=True,
        deafen_members=True,
        move_members=True,
        use_voice_activation=True,
        change_nickname=True,
        manage_nicknames=False,
        manage_roles=False,
        manage_permissions=False,
        manage_webhooks=False,
        manage_emojis=False,
    )

    DEFAULT_ASSOCIATE_PERMISSIONS = PermissionOverwrite(
        create_instant_invite=True,
        kick_members=False,
        ban_members=False,
        administrator=False,
        manage_channels=False,
        manage_guild=False,
        add_reactions=True,
        view_audit_log=True,
        priority_speaker=False,
        stream=True,
        read_messages=True,
        view_channel=True,
        send_messages=True,
        send_tts_messages=True,
        manage_messages=True,
        embed_links=True,
        attach_files=True,
        read_message_history=True,
        mention_everyone=True,
        external_emojis=True,
        use_external_emojis=True,
        view_guild_insights=False,
        connect=True,
        speak=True,
        mute_members=True,
        deafen_members=False,
        move_members=False,
        use_voice_activation=True,
        change_nickname=True,
        manage_nicknames=False,
        manage_roles=False,
        manage_permissions=False,
        manage_webhooks=False,
        manage_emojis=False,
    )

    DEFAULT_FORMER_PERMISSIONS = DEFAULT_OTHERS_PERMISSIONS


@dataclass()
class CreateOrganization(GuildCommand):
    name: str = None
    admin_text_rooms = ['management', ]
    admin_voice_rooms = ['Dome', ]
    private_text_rooms = ['general', ]
    private_voice_rooms = ['Main Room', ]
    friend_text_rooms = ['coffee', ]
    friend_voice_rooms = ['Coffee', 'Rest Room', ]
    open_text_rooms = []
    open_voice_rooms = ['Living Room', ]

    default_admin_permissions: PermissionOverwrite = field(default_factory=PermissionOverwrite)
    default_associate_permissions: PermissionOverwrite = field(default_factory=PermissionOverwrite)
    default_others_permissions: PermissionOverwrite = field(default_factory=PermissionOverwrite)
    default_former_permissions: PermissionOverwrite = field(default_factory=PermissionOverwrite)

    @async_batch
    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> Result:
        admin_role = await CreateRole(
            name=f'{self.name}\'s Admin',
        ).do_on_guild(client, guild)
        yield admin_role
        associate_role = await CreateRole(
            name=f'{self.name}\'s Associate',
            hoist=True,
        ).do_on_guild(client, guild)
        yield associate_role
        former_role = await CreateRole(
            name=f'{self.name}\'s Former Associate',
        ).do_on_guild(client, guild)
        yield former_role
        created_category = await CreateCategory(
            name=self.name,
            overwrites={
                admin_role.role: self.default_admin_permissions,
                associate_role.role: self.default_associate_permissions,
                former_role.role: self.default_former_permissions,
                DEFAULT_ROLE: self.default_others_permissions,
            }
        ).do_on_guild(client, guild)
        yield created_category

        for room_name in self.admin_voice_rooms:
            yield await CreateVoiceChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    associate_role.role: self.default_others_permissions,
                    former_role.role: NO_PERMISSIONS,
                    DEFAULT_ROLE: NO_PERMISSIONS,
                }
            ).do_on_guild(client, guild)
        for room_name in self.admin_text_rooms:
            yield await CreateTextChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    associate_role.role: self.default_others_permissions,
                    former_role.role: NO_PERMISSIONS,
                    DEFAULT_ROLE: NO_PERMISSIONS,
                }
            ).do_on_guild(client, guild)

        for room_name in self.private_voice_rooms:
            yield await CreateVoiceChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    associate_role.role: self.default_associate_permissions,
                    former_role.role: self.default_former_permissions,
                    DEFAULT_ROLE: self.default_others_permissions,
                }
            ).do_on_guild(client, guild)
        for room_name in self.private_text_rooms:
            yield await CreateTextChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    associate_role.role: self.default_associate_permissions,
                    former_role.role: self.default_former_permissions,
                    DEFAULT_ROLE: self.default_others_permissions,
                }
            ).do_on_guild(client, guild)

        for room_name in self.friend_voice_rooms:
            yield await CreateVoiceChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    associate_role.role: self.default_associate_permissions,
                    former_role.role: self.default_associate_permissions,
                    DEFAULT_ROLE: self.default_others_permissions,
                }
            ).do_on_guild(client, guild)
        for room_name in self.friend_text_rooms:
            yield await CreateTextChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    associate_role.role: self.default_associate_permissions,
                    former_role.role: self.default_associate_permissions,
                    DEFAULT_ROLE: self.default_others_permissions,
                }
            ).do_on_guild(client, guild)

        for room_name in self.open_voice_rooms:
            yield await CreateVoiceChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    DEFAULT_ROLE: self.default_associate_permissions,
                }
            ).do_on_guild(client, guild)
        for room_name in self.open_text_rooms:
            yield await CreateTextChannel(
                name=room_name,
                category=created_category.category,
                overwrites={
                    admin_role.role: self.default_admin_permissions,
                    DEFAULT_ROLE: self.default_associate_permissions,
                }
            ).do_on_guild(client, guild)

    @classmethod
    def open_organization(cls, name: str, **kwargs) -> 'CreateOrganization':
        return cls(
            name=name,
            **kwargs,
            default_admin_permissions=OpenOrganization.DEFAULT_ADMIN_PERMISSIONS,
            default_associate_permissions=OpenOrganization.DEFAULT_ASSOCIATE_PERMISSIONS,
            default_former_permissions=OpenOrganization.DEFAULT_FORMER_PERMISSIONS,
            default_others_permissions=OpenOrganization.DEFAULT_OTHERS_PERMISSIONS,
        )

    @classmethod
    def private_organization(cls, name: str, **kwargs) -> 'CreateOrganization':
        return cls(
            name=name,
            **kwargs,
            default_admin_permissions=PrivateOrganization.DEFAULT_ADMIN_PERMISSIONS,
            default_associate_permissions=PrivateOrganization.DEFAULT_ASSOCIATE_PERMISSIONS,
            default_former_permissions=PrivateOrganization.DEFAULT_FORMER_PERMISSIONS,
            default_others_permissions=PrivateOrganization.DEFAULT_OTHERS_PERMISSIONS,
        )



from dataclasses import dataclass
from typing import Callable, TypeVar, Generic, Dict, Iterable, Optional, List

E = TypeVar('E')


class Predicate(Generic[E]):

    def __call__(self, e: E) -> bool:
        raise NotImplementedError()

    def filter(self, iterable: Iterable[E]) -> Iterable[E]:
        for e in iterable:
            if self(e):
                yield e

    def any(self, iterable: Iterable[E]) -> Optional[E]:
        for e in iterable:
            if self(e):
                return e
        return None

    def __neg__(self) -> 'Predicate':
        return NotPredicate(self)

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        if other is None:
            return False
        if self.__class__ != other.__class__:
            return False
        return self.__dict__ == other.__dict__

    def __or__(self, other: 'Predicate') -> 'Predicate':
        return OrPredicate([self, other])

    def __and__(self, other: 'Predicate') -> 'Predicate':
        return AndPredicate([self, other])


@dataclass()
class FnPredicate(Predicate[E]):
    fn: Callable[[E], bool]

    def __call__(self, e: E) -> bool:
        return self.fn(e)


@dataclass()
class OrPredicate(Predicate[E]):
    predicates: List[Predicate[E]]

    def __call__(self, e: E) -> bool:
        for predicate in self.predicates:
            if predicate(e):
                return True
        return False

    def __or__(self, other: 'Predicate') -> 'Predicate':
        return OrPredicate(self.predicates + [other])


@dataclass()
class AndPredicate(Predicate[E]):
    predicates: List[Predicate[E]]

    def __call__(self, e: E) -> bool:
        for predicate in self.predicates:
            if not predicate(e):
                return False
        return True

    def __and__(self, other: 'Predicate') -> 'Predicate':
        return AndPredicate(self.predicates + [other])


@dataclass()
class NotPredicate(Predicate[E]):
    predicate: Predicate[E]

    def __call__(self, e: E) -> bool:
        return not self.predicate(e)

    def __not__(self) -> 'Predicate':
        return self.predicate


class AnyPredicate(Predicate):

    def __call__(self, *args, **kwargs) -> bool:
        return True

    def __not__(self) -> 'Predicate':
        return NONE


class NonePredicate(Predicate):
    def __call__(self, *args, **kwargs) -> bool:
        return False

    def filter(self, iterable: Iterable[E]) -> Iterable[E]:
        return []

    def any(self, iterable: Iterable[E]) -> Optional[E]:
        return None

    def __not__(self) -> 'Predicate':
        return ANY


ANY = AnyPredicate()

NONE = NonePredicate()


@dataclass()
class ByAttributesPredicate(Predicate[E]):
    attr: Dict

    def __call__(self, e: E) -> bool:
        for k, v in self.attr.items():
            if getattr(e, k) != v:
                return False
        return True

    @classmethod
    def create(cls, **attr):
        return cls(attr)

    @classmethod
    def parse(cls, query: str):
        return cls()

    def __hash__(self):
        return id(self)



from discord import Client
from dataclasses import dataclass
from discord.user import User
from discord.message import Message

from discord_coworking.bot.api import MessageHandler
from discord_coworking.bot.handler import MessageHandlerChain
from discord_coworking.command.predicate import Predicate, E


@dataclass()
class NotMySelf(Predicate[Message]):
    client: Client

    def __call__(self, e: Message) -> bool:
        return e.author.id != self.client.user.id


@dataclass()
class StartByMentioningMe(Predicate[Message]):
    client: Client

    def __call__(self, e: Message) -> bool:
        mention = f'<@!{self.client.user.id}>'
        return e.content.startswith(mention)


class CoworkingBot(Client):
    message_handler: MessageHandlerChain

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message_handler = MessageHandlerChain(
            handlers=[],
            predicate=NotMySelf(self),
        )

    async def on_message(self, message: Message):
        await self.message_handler(self, message)



import os

from setuptools import setup, find_packages

__version__ = '0.0.dev1666029291'


def read(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name)) as f:
        return f.read()


setup(
    name='discord-coworking',
    version=__version__,
    author='Ettore Leandro Tognoli',
    author_email='ettoreleandrotognoli@gmail.com',
    license='Apache License 2.0',
    data_files=[
        'LICENSE',
    ],
    description='Discord coworking server management toolkit',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    packages=find_packages(
        './src/main/python/',
    ),
    package_dir={'': 'src/main/python'},
    include_package_data=True,
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
    ],
    python_requires='>=3.10',
    install_requires=[
        'discord.py==1.7.*',
        'emoji==0.6.*',
        'docker==6.*',
    ],
    tests_require=[
        'coverage',
        'pylint',
    ],
    entry_points={
        'console_scripts': [
            'discord-coworking=discord_coworking.__main__'
        ]
    },
)



import logging
from dataclasses import dataclass, field
from logging import Logger

from discord import Client as DiscordClient
from discord import Guild

from discord_coworking.command.api import Command, Result
from discord_coworking.command.decorator import async_batch
from discord_coworking.command.predicate import ByAttributesPredicate, Predicate, ANY

__logger__: Logger = logging.getLogger(__name__)


class GuildPredicate(ByAttributesPredicate[Guild]):

    @classmethod
    def of(cls, guild: Guild):
        return cls.create(id=guild.id)


@dataclass()
class GuildResult(Result):
    guild: Predicate[Guild] = ANY

    async def undo(self, client: DiscordClient):
        for guild in self.guild.filter(client.guilds):
            await self.undo_on_guild(client, guild)

    async def undo_on_guild(self, client: DiscordClient, guild: Guild):
        raise NotImplementedError()


@dataclass()
class GuildCommand(Command):
    guild: Predicate[Guild] = ANY
    logger: Logger = field(default=__logger__)

    @async_batch
    async def do(self, client: DiscordClient) -> Result:
        for guild in self.guild.filter(client.guilds):
            self.logger.info(f'Running command {self} on guild {guild.name}({guild.id})')
            yield await self.do_on_guild(client, guild)

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> Result:
        raise NotImplementedError()
